#!/bin/bash

################################################################################
#
#                   Cyberwatch Application Launching Script
#
################################################################################

set -o pipefail

kubernetes_namespace="${KUBERNETE_NAMESPACE:-cyberwatch}"
helm_release="${helm_release:-${kubernetes_namespace}}"
container_repository="${CBW_CONTAINER_REPOSITORY:-${container_registry}}"
helm_repository="${CBW_HELM_REPOSITORY:-${container_repository}}"
helm_chart="${CBW_HELM_CHART:-cyberwatch-chart}"
helm_chart_cache_dir="${CBW_HELM_CHART_CACHE_DIR:-$HOME/.cache/cyberwatch}"
secret_file="${CBW_VALUES_FILE:-$config_dir/secrets.yml}"
containers_file="${CBW_VALUES_FILE:-$config_dir/containers.yml}"

# Aliases
KUBE="${KUBE:-microk8s kubectl -n $kubernetes_namespace}"
HELM="${HELM:-microk8s helm3 -n $kubernetes_namespace}"

# Variables defined in `cyberwatch` script. Ajouté originellement pour faire taire SC2154
container_registry="${container_registry}"
exec_interactive="${exec_interactive}"
is_offline="${is_offline}"
local_exposed_ip="${local_exposed_ip}"
no_db="${no_db}"
no_redis="${no_redis}"
node_type="${node_type}"
cert_dir="${cert_dir}"
sshkey_dir="${sshkey_dir}"

configure_sidekiq_ressources() { return; }

extra_help() {
  cat <<EOF
    describe [SERVICE]               Describe the state of a pod.
EOF
}

## CONFIGURE

generate_new_secrets() {
  install -m 640 /dev/null "${secret_file}"
  cat <<-EOF >"${secret_file}"
database:
  password: "${MYSQL_PASSWORD:-$(openssl rand -hex 16)}"
  root_password: "${MYSQL_ROOT_PASSWORD:-$(openssl rand -hex 16)}"

redis:
  password: "${REDIS_PASSWORD:-$(openssl rand -hex 16)}"

key:
  base: "${SECRET_KEY_BASE:-$(openssl rand -hex 64)}"
  credential: "${SECRET_KEY_CREDENTIAL:-$(openssl rand -hex 64)}"
EOF
}

add_absent_secrets() {
  true # Do nothing
}

configure_orchestrator() {
  # Create namespace
  $KUBE get ns "$kubernetes_namespace" >/dev/null 2>&1 ||
    $KUBE create namespace "$kubernetes_namespace" 2>&1 ||
    fail "Failure during namespace creation"

  microk8s status >/dev/null || microk8s start
  if [ "$is_offline" == "true" ]; then
     microk8s enable dns >/dev/null
     $HELM version >/dev/null 2>&1 || \
       error "Warning : Helm doesn't seem to be installed or environnement variable HELM is misconfigured."
  else
    microk8s enable dns helm3 >/dev/null
  fi

  # The file containing Helm credentials was changed from ~/.config/helm/registry.json to
  # ~/.config/helm/registry/config.json in version 3.8.0 of Helm to align with Docker. However, this is not
  # retrocompatible. Because some clients still use Helm v3.5.2, and may migrate to 3.8.0, let's copy the credentials so
  # that the password will not be asked again.
  # https://github.com/helm/helm/commit/e08a9270be1b3be83397c73e7635fd2ecd8f043d
  if [ -f ~/.config/helm/registry.json ] && ! [ -f ~/.config/helm/registry/config.json ]; then
    mkdir -p ~/.config/helm/registry
    cp ~/.config/helm/registry.json ~/.config/helm/registry/config.json
  fi

  # Login to container registry
  if {
    [ "$($KUBE get secret cyberwatch-credentials 2>/dev/null | wc -l)" -lt 2 ] || \
      ! grep -q "$helm_repository" ~/.config/helm/registry/config.json 2>/dev/null;
  } &&
    ! is_offline; then

    authentified="false"
    while [ "${authentified}" = "false" ]; do
      info "Cyberwatch's container repository credentials not found locally."
      read -rp 'Cyberwatch repository username: ' cbw_username
      read -srp 'Cyberwatch repository password: ' cbw_password
      echo
      echo "${cbw_password}" | HELM_EXPERIMENTAL_OCI=1 $HELM registry login "${helm_repository}" \
        --username "cbw\$${cbw_username#cbw\$}" \
        --password-stdin 2>&1 | hide_kubectl_warning
      if ! [ "${PIPESTATUS[1]}" -eq 0 ]; then
        error "Wrong credentials"
        continue
      fi
      $KUBE delete secret cyberwatch-credentials 2>&1 | grep -vE "deleted|not found" || true
      $KUBE create secret docker-registry cyberwatch-credentials \
        --docker-server="$container_repository" \
        --docker-username "cbw\$${cbw_username#cbw\$}" \
        --docker-password "$cbw_password" || { error "Failure while storing credentials" && continue; }
      authentified="true"
    done
  else
    info "Cyberwatch's container repository credentials found, skipping authentication"
  fi
}

create_pki_secrets() {
  (
    cd "$cert_dir" || exit
    $KUBE get secret cbw-root-ca-cert >/dev/null 2>&1 || \
      $KUBE create secret generic cbw-root-ca-cert --from-file=cbw-root-ca-cert.pem

    if [ "$node_type" != "satellite" ]; then
    $KUBE get secret cbw-db-tls >/dev/null 2>&1 || \
      $KUBE create secret tls cbw-db-tls --cert=cbw-db-cert.pem --key=cbw-db-key.pem
    $KUBE get secret cbw-redis-tls >/dev/null 2>&1 || \
      $KUBE create secret tls cbw-redis-tls --cert=cbw-redis-cert.pem --key=cbw-redis-key.pem
    $KUBE get configmap cyberwatch-tls >/dev/null 2>&1 || \
      $KUBE create configmap cyberwatch-tls --from-file=/usr/share/cyberwatch/99-cyberwatch-tls.cnf
    fi
  )
}

relink_containers_config() { return; }

generate_new_containers_file() {
  cat <<-EOF >"${containers_file}"
# This file override some values of the helm chart
# All available fields are listed and documented in file ${containers_file}.example

EOF

  if [ "$is_offline" == "true" ]; then
    cat <<-EOF >>"${containers_file}"
global:
  image:
    pullPolicy: Never

EOF
  fi

  cat <<-EOF >>"${containers_file}"
node:
  name: $(hostname)
  ip: ${local_exposed_ip}
  type: ${node_type:-single}

nginx:
  resolver: $(microk8s kubectl get svc kube-dns --namespace=kube-system -o=jsonpath='{.spec.clusterIP}')

EOF

  if [ "$is_offline" == "true" ]; then
    cat <<-EOF >>"${containers_file}"
sidekiq:
  extraVolumes:
  - name: data-securitydb
    hostPath:
      path: /var/lib/volumes/cyberwatch/securitydb
  extraVolumeMounts:
  - name: data-securitydb
    mountPath: /security_database

EOF
  fi

  cat <<-EOF >>"${containers_file}"
database:
  hostname: "${external_database:-${master_host:-db}}"
  name: "${MYSQL_DATABASE:-olympe}"
  user: "${MYSQL_USER:-olympe}"
  port: "${MYSQL_PORT:-3306}"

  ## Set to true to disable the cluster database and use an external one
  external: $({ [ "${node_type}" == "satellite" ] || [ "${no_db}" == "true" ]; } && echo true || echo false)

EOF

  cat <<-EOF >>"${containers_file}"
redis:
  host: "${master_host:-redis}"
  user: "${REDIS_USER:-default}"
  port: "${REDIS_PORT:-6379}"

  ## Set to true to disable the cluster redis server and user an external one
  external: $({ [ "${node_type}" == "satellite" ] || [ "${no_redis}" == "true" ]; } && echo true || echo false)

EOF

  # The use of the thirdParties container is not required when the Cyberwatch
  # instance has access to the internet because updates can be downloaded. To speed
  # up the launch of Cyberwatch, we instruct not to use this container. It is all
  # the more effective because of the large size of the container.
  if [ "${is_offline}" != "true" ]; then
    cat <<-EOF >>"${containers_file}"
thirdParties:
  enabled: false

EOF
  fi

  if [ "${node_type}" == "satellite" ]; then
    cat <<-EOF >>"${containers_file}"
elasticsearch:
  enabled: false

kibana:
  enabled: false

EOF
  fi

}

start_certbot() {
  fail "Not Implemented"
}

## OPERATIONAL

start_instance() {
  argument=$1
  case $argument in
  "--no-progress")
    NO_PROGRESS="true"
    ;;
  esac

  install -m 750 -d /var/lib/volumes "/var/lib/volumes/${helm_release}"

  identify_chart_address
  generate_local_api_token

  CBW_ORCHESTRATOR_VERSION=$(snap list microk8s | sed '1d')
  $HELM upgrade -i "$helm_release" "$chart_address" --values "${containers_file}" \
    --values "${secret_file}" \
    --set web.env.CBW_ON_PREMISE_VERSION="$CBW_ON_PREMISE_VERSION" \
    --set sidekiq.env.CBW_ON_PREMISE_VERSION="$CBW_ON_PREMISE_VERSION" \
    --set sidekiqNode.env.CBW_ON_PREMISE_VERSION="$CBW_ON_PREMISE_VERSION" \
    --set sidekiqMaster.env.CBW_ON_PREMISE_VERSION="$CBW_ON_PREMISE_VERSION" \
    --set web.env.CBW_ORCHESTRATOR_VERSION="$CBW_ORCHESTRATOR_VERSION" \
    --set web.env.LOCAL_API_TOKEN="$(cbw_get-local-api-token)" \
    --set web.env.BACKUP_DATE="${BACKUP_DATE}" \
    2>&1 | hide_kubectl_warning || fail "Failure while deploying Helm Chart"
  echo "# DO NOT MODIFY THIS FILE. THIS IS AN EXAMPLE FILE ONLY. IT MAY BE OVERWRITTEN." > "${containers_file}.example"
  $HELM show values "$chart_address" >>"${containers_file}".example 2>&1 | hide_kubectl_warning

  # Pulling image if user is sudoers without password prompt
  if [ -z "${is_offline}" ] && ! [ "${NO_PROGRESS}" == "true" ] && sudo -n true >/dev/null 2>&1; then
    CREDS=$(
      microk8s kubectl get secret \
        -n cyberwatch cyberwatch-credentials \
        -o 'jsonpath={.data.\.dockerconfigjson}' |
        base64 -d | awk -F'"' '{ printf("%s:%s", $8, $12) }'
    )
    microk8s ctr i pull "${container_repository}/olympe:latest" -u "$CREDS"
  fi
}

identify_chart_address() {
  is_offline && { chart_address="${helm_chart_cache_dir}/${helm_chart}"; return; }

  if $HELM chart --help >/dev/null 2>&1; then
    fail "Your version of Helm is deprecated. Please use a version more recent that 3.8.0.
If Helm is packaged with Microk8s (the default), please ensure Microk8s version is at least 1.24.0."
  else
    chart_address="oci://${helm_repository}/${helm_chart}"
  fi
}

stop_instance() {
  $HELM uninstall "$helm_release" 2>&1 | hide_kubectl_warning | grep -v "not found"
  info "Cyberwatch is deleting resources in the background."
  info "If you want to watch the states of the pods, run"
  echo -e "\n\twatch $0 status\n"
}

cbw_restart() {
  LONG=skip-backup,verify
  OPTS=$(getopt --longoptions $LONG -- "$0" "$@")

  eval set -- "$OPTS"

  while true; do
      case "$1" in
          "--skip-backup")
              export skip_backup=true
              shift
              ;;
          "--verify")
              verify=true
              shift
              ;;
          --)
              shift;
              break
              ;;
          *)
              echo "Unexpected option: $1"
          ;;
      esac
  done

  [ -f "$containers_file" ] || fail "File $containers_file does not exist. Did you run '$0 configure' ?"

  info "Restarting containers"
  is_satellite || backup_db_if_no_recent
  [ -d "${sshkey_dir}" ] || configure_ssh_key
  is_satellite || create_pki_secrets
  if ! is_offline && [ "$verify" = "true" ]; then
      cbw_verify-images-signature || fail "Failed to verify the signature of images"
      success "The signature of the images is valid."
  fi
  cbw_stop
  echo -n "Waiting for volumes to be detatched"
  while [ "$($KUBE get pvc 2>/dev/null | wc -l)" -gt "1" ]; do
    echo -n .
    sleep 1
  done
  echo
  cbw_start
  info "Cyberwatch is reloading resources in the background."
  info "If you want to watch the states of the pods, run"
  echo -e "\n\t$0 status\n"
}

load_ssh_secret() {
    $KUBE delete secret ssh-private-key web-scanner-ssh-authorized-keys >/dev/null 2>&1
    $KUBE create secret generic ssh-private-key --from-file="${sshkey_dir}/id_ed25519"
    $KUBE create secret generic web-scanner-ssh-authorized-keys --from-file=authorized_keys="${sshkey_dir}/id_ed25519.pub"
}

cbw_status() {
  $KUBE get po
}

# Select a pod for service $1.
# Supports adding extra fields with arguments starting from $2
# Extra kubernetes filters can be requested with $3 (ex:  --field-selector=status.phase=Running)
# Store the selected pod name or field in global variable 'selected_pod'
select_pod_for_service() {
  [ -z "$1" ] && { cbw_help; fail "Please specify a service name."; }
  # shellcheck disable=SC2086
  pods="$($KUBE get po -l app="$1" -o name $3)"
  if [ "$pods" = "" ]; then
    if [ "$3" = "" ]; then
      fail "No pod match '$1'"
    else
      fail "No pod match '$1' with filter '$3'"
    fi
  fi

  if [ "$(echo "$pods" | wc -w)" -gt 1 ]; then
    echo "Multiple containers match the required service. Please select one:"
    export PS3="Choice: "; select pod in $pods $2; do
      selected_pod="$pod"
      break
    done
  else
    selected_pod="$pods"
  fi
}

# Display logs of a pod. This can be used with the '-f' flag.
cbw_logs() {
  select_pod_for_service "$1" "all" "--field-selector=status.phase=Running"
  if [ "$selected_pod" == "all" ]; then
    $KUBE logs -l app="$1" --tail=-1 "${@:2}"
  else
    $KUBE logs "$selected_pod" "${@:2}"
  fi
}

# Describe a pod.
cbw_describe() {
  select_pod_for_service "$1" "all"
  if [ "$selected_pod" == "all" ]; then
    $KUBE describe po -l app="$1" "${@:2}"
  else
    $KUBE describe "$selected_pod" "${@:2}"
  fi
}

# Execute a command in a pod. Default to using '/bin/sh'.
cbw_exec() {
  ARGS="${*:2}"
  select_pod_for_service "$1" "" "--field-selector=status.phase=Running"
  PODS="$selected_pod"
  if [ -z "$PODS" ]; then
    echo "Pod $1 is not started."
    return 1
  fi
  if [ "$exec_interactive" == "false" ]; then
    $KUBE exec -i "$PODS" -- sh -c "${ARGS:-/bin/sh}"
  else
    $KUBE exec -it "$PODS" -- sh -c "${ARGS:-/bin/sh}"
  fi
}

cbw_show-config() {
  identify_chart_address

  $HELM template "$helm_release" "$chart_address" --values "${containers_file}" \
    --values "${secret_file}" \
    --set web.env.CBW_ON_PREMISE_VERSION="$CBW_ON_PREMISE_VERSION" \
    --set sidekiq.env.CBW_ON_PREMISE_VERSION="$CBW_ON_PREMISE_VERSION" \
    --set sidekiqNode.env.CBW_ON_PREMISE_VERSION="$CBW_ON_PREMISE_VERSION" \
    --set sidekiqMaster.env.CBW_ON_PREMISE_VERSION="$CBW_ON_PREMISE_VERSION"
}

check_orchestrator() {
    if command -v microk8s >/dev/null; then
        microk8s_version=$(snap list microk8s | grep microk8s | awk '{ printf("%s", $2) }')
        microk8s_channel=$(snap list microk8s | grep microk8s | awk '{ printf("%s", $4) }')

        success "  > Microk8s version $microk8s_version (channel $microk8s_channel)"
        if echo "$microk8s_version" | grep -qE "v1\.2[0123]\."; then
          warn "    > Microk8s version ($microk8s_version) is not up-to-date. Please consider updating to a version more recent that v1.24.0."
        fi
        if [ "$microk8s_channel" != "latest/stable" ]; then
          warn "    > Microk8s snap package channel is $microk8s_channel. Please consider using 'latest/stable' instead."
        fi
    elif [ -f /snap/bin/microk8s ]; then
         error "  > Microk8s is installed but not present in PATH."
         error "  > Please edit this variable globally (/etc/environnement) or locally."
    fi

    helm_version=$($HELM version --short 2>/dev/null)
    if echo "$helm_version" | grep -qE "v3\.[34567]\."; then
      warn "  > Helm version $helm_version"
      warn "    > Helm version $helm_version is deprecated, please update to a version more recent than 3.8.0."
      if echo "$HELM" | grep -q 'microk8s'; then
        warn "    > It seems that Helm is packaged with Microk8s. Please ensure Microk8s version is at least v1.24.0."
        warn "    > Then update Helm with:"
        warn "    > microk8s disable helm3; microk8s enable helm3"
      fi
    else
      success "  > Helm version $helm_version"
    fi
}

check_container_status() {
    containers=$($KUBE get pods | sed '1d')
    if echo "$containers" | grep -q 'Running'; then
        if echo "$containers" | grep -q -v 'Running'; then
            error "  > Found non running container(s). Use '${0##*/} describe' to inspect failed pods:"
            echo -e "\n$containers\n" | grep -v 'Running'
        else
            success "  > All container are running"
        fi
    else
        error "  > Cyberwatch is started but no running container found"
        echo
        echo -e "\n$containers\n" | grep -v 'Running'
        echo
    fi
}

# Hide the warning from kubectl:
# "WARNING: Kubernetes configuration file is group-readable. This is insecure. Location: /var/snap/microk8s/2262/credentials/client.config"
#
# This is normal with microk8s as the cyberwatch user need to be in the `microk8s` group
hide_kubectl_warning() {
  grep -v "WARNING: Kubernetes configuration file is group-readable." -
}

report_orchestrator_config() {
    report_file_content "$config_dir/containers.yml"
}

report_orchestrator_version() {
    report_command "$HELM version"
    report_command "snap list microk8s"
}

report_orchestrator_containers() {
    report_command "$HELM list"
    report_command "$KUBE describe pods,pvc,pv,service,secrets"
    report_command "microk8s ctr i ls"
}

# Utils

is_satellite() {
  grep "type: satellite" "${containers_file}" -q
}

is_offline() { [ "$is_offline" == "true" ]; }

is_running() { $HELM list 2>&1 | grep -q "$helm_release"; }
