#!/bin/bash

################################################################################
#
#                   Cyberwatch Command Line Interface
#
################################################################################

set -o pipefail

swarm_stack_name="${CBW_SWARM_STACK_NAME:-cbwonpremise}"

available_config_dir="${CBW_AVAILABLE_CONFIG_DIR:-$config_dir/configs-available}"
enabled_config_dir="${CBW_ENABLED_CONFIG_DIR:-$config_dir/configs-enabled}"

secret_file="${CBW_SECRET_FILE:-$config_dir/secrets.env}"
containers_file="${CBW_CONTAINER_ENV_FILE:-$config_dir/containers.env}"

# shellcheck disable=SC2269 # Variables defined in `cyberwatch` script. Ajouté originellement pour faire taire SC2154
certbot="${certbot}"
certbot_domain_name="${certbot_domain_name}"
config_file="${config_file}"
container_registry="${container_registry}"
exec_interactive="${exec_interactive}"
is_offline="${is_offline}"
no_db="${no_db}"
no_redis="${no_redis}"
node_type="${node_type}"
sshkey_dir="${sshkey_dir}"
image_cert="${image_cert}"
network_container_scanner="cbwonpremise_network_container_scanner"

# Define if docker must pull new images. If set to false (default), docker will
# pull new images when using cbw_start and cbw_restart. If set to true (with
# --no-pull option on cbw_restart), no image will be pulled.
no_pull="${no_pull}"

extra_help() { return; }
create_pki_secrets() { return; }

## CONFIGURE

configure_orchestrator() {
    ln -sf -t "$config_dir" /usr/share/cyberwatch/swarm/configs-available

    systemctl start docker

    if [ "$(docker info --format '{{.Swarm.LocalNodeState}}')" != "active" ]; then
        docker swarm init --advertise-addr="$(hostname -I | cut -d' ' -f1)" >/dev/null
    fi

    if ! grep -q "${container_registry%%/*}" ~/.docker/config.json 2>/dev/null && ! is_offline; then
        echo "Please log in to the container registry $container_registry"
        read -rp 'Cyberwatch repository username: ' cbw_username
        read -srp 'Cyberwatch repository password: ' cbw_password
        echo
        [ "$container_registry" == "harbor.cyberwatch.fr/cbw-on-premise" ] && cbw_username="cbw\$${cbw_username#cbw\$}"
        echo "$cbw_password" | docker login "$container_registry" --password-stdin -u "$cbw_username" || fail "Cannot access to Cyberwatch Docker Registry"
    fi
}

# Check if machine has more than 4 CPUs
# If yes, enable 2 sidekiq replicas.
# Else, use the default number of sidekiq.
configure_sidekiq_ressources() {
    cpu_number=$(($(lscpu -eCPU | wc -l) - 1)) # minus 1 because there is a header.
    sidekiq_replicas=$(awk -F= '/^CBW_SIDEKIQ_REPLICAS/ { print $2 }' "${config_file}")

    if [[ ${cpu_number} -ge 4 ]] && [[ sidekiq_replicas -lt 2 ]]; then
        sed -i -E 's/^CBW_SIDEKIQ_REPLICAS=[0-9]+/CBW_SIDEKIQ_REPLICAS=2/' "${config_file}"
        info "Set 2 Sidekiq replicas instead of ${sidekiq_replicas} in ${config_file} because more than 4 CPUs are available."
    fi
}


generate_new_secrets() {
    install -m 640 /dev/null "${secret_file}"
    cat <<-EOF >"${secret_file}"
MYSQL_ROOT_PASSWORD=$(openssl rand -hex 16)
MYSQL_PASSWORD=$(openssl rand -hex 16)
REDIS_PASSWORD=$(openssl rand -hex 16)
SECRET_KEY_BASE=$(openssl rand -hex 64)
SECRET_KEY_CREDENTIAL=$(openssl rand -hex 64)
EOF
}

add_absent_secrets() {
    grep -q REDIS_PASSWORD "${secret_file}" || echo "REDIS_PASSWORD=$(openssl rand -hex 16)" >> "${secret_file}"
}

relink_containers_config() {
    mkdir -p "${enabled_config_dir}/compose-service"
    enable_config 10-base.yml

    [ "$node_type" != "satellite" ]
    enable_config 11-sidekiq_master.yml

    [ "$is_offline" == "true" ]
    enable_config 12-third-parties.yml

    true # Always
    enable_config 35-web-scanner.yml

    enable_config compose-service/container-scanner.yml

    [ "$node_type" == "satellite" ]
    enable_config 15-satellite.yml

    [ "$node_type" != "satellite" ] && [ "$no_redis" != "true" ]
    enable_config 20-redis.yml

    [ "$node_type" != "satellite" ]
    enable_config 30-elk.yml

    [ "$node_type" != "satellite" ] && [ "$no_db" != "true" ]
    enable_config 40-db.yml

    [ "$node_type" == "master" ] && [ "$no_db" != "true" ]
    enable_config 50-expose-db.yml

    [ "$node_type" == "master" ] && [ "$no_redis" != "true" ]
    enable_config 51-expose-redis.yml

    [ "$certbot" == "true" ]
    {
        enable_config 60-certbot.yml
        mkdir -p "${config_dir}"/letsencrypt
    }

    [ "$is_offline" == "true" ]
    enable_config 70-offline.yml
    is_offline && mkdir -p /var/lib/cyberwatch/security_database
}

migrate_version() {
    # Function used to migrate files between versions of cbw-on-premise.
    echo "Start migration process"
    if [ -d "${enabled_config_dir}" ]; then
        ## Migration of compose file for cbw-on-premise 5.21
        if [ ! -L "${enabled_config_dir}/11-sidekiq_master.yml" ] && [ ! -L "${enabled_config_dir}/15-satellite.yml" ]; then
            info "Create symlink for '${available_config_dir}/11-sidekiq_master.yml' -> '${enabled_config_dir}/11-sidekiq_master.yml'"
            ln -sr "${available_config_dir}/11-sidekiq_master.yml" "${enabled_config_dir}/11-sidekiq_master.yml"
        fi
        ## Remove symbolic link for 11-cron.yml that was removed in cbw-on-premise 5.21
        ## Allows the command to fail with `|| true` because the file is not present for next updates.
        if [ -L "${enabled_config_dir}/11-cron.yml" ]; then
            info "Removing symlink for '${enabled_config_dir}/11-cron.yml'"
            rm "${enabled_config_dir}/11-cron.yml"
        fi
        ## Creation of the compose-service directory for cbw-on-premise 5.24.
        if [ ! -d "${enabled_config_dir}/compose-service" ]; then
            info "Creation of the directory '${enabled_config_dir}/compose-service' for storing Compose service files."
            mkdir -p "${enabled_config_dir}/compose-service"
        fi
        ## Migration of compose-file (container-scanner) for cbw-on-premise 5.24.
        if [ ! -L "${enabled_config_dir}/compose-service/container-scanner.yml" ]; then
            info "Create symlink for '${available_config_dir}/compose-service/container-scanner.yml' -> '${enabled_config_dir}/compose-service/container-scanner.yml'"
            ln -sr "${available_config_dir}/compose-service/container-scanner.yml" "${enabled_config_dir}/compose-service/container-scanner.yml"
        fi
    fi

    echo "Stop migration process"
}

enable_config() {
    should_remove=$?
    if [ "$should_remove" == "1" ]; then
        [ -f "${enabled_config_dir}/$1" ] || return # File does not exist
        echo "Deleting ${enabled_config_dir}/$1"
        rm "${enabled_config_dir}/$1"
    else
        [ -f "${available_config_dir}/$1" ] || {
            echo "Error: config file ${available_config_dir}/$1 not found"
            exit 1
        }
        [ -f "${enabled_config_dir}/$1" ] && return # Config already enabled
        echo "Enabling $1"
        ln -s -r "${available_config_dir}/$1" "${enabled_config_dir}/$1"
    fi
}

generate_new_containers_file() {
    cat <<-EOF >"${containers_file}"
MYSQL_HOSTNAME=${external_database:-${master_host:-db}}
MYSQL_DATABASE=olympe
MYSQL_USER=olympe
MYSQL_PORT=3306
REDIS_HOST=${external_redis_server:-${master_host:-redis}}
REDIS_PORT=6379
REDIS_USER=default
CERTBOT_NGINX_DOMAIN=$certbot_domain_name
ENABLE_HSTS='false'
ONLY_API_AND_THIRD_PARTIES='false'
EOF
}

start_certbot() {
    for _ in $(seq 1 10); do
        certbot_container=$(docker ps -q --filter="NAME=${swarm_stack_name}_certbot")
        [ -n "$certbot_container" ] && break
        echo "Waiting for certbot container to be running"
        sleep 3
    done
    cbw_exec certbot certbot certonly --authenticator webroot --non-interactive --agree-tos --webroot-path /var/www/certbot --register-unsafely-without-email -d "${CERTBOT_NGINX_DOMAIN:-$certbot_domain_name}"
}

## OPERATIONAL

start_instance() {
    migrate_compose
    generate_local_api_token
    load_secrets
    read -ra config_files < <(find "${enabled_config_dir}" -maxdepth 1 -name "*.yml" -print0 | sort -z | xargs -0 -n 1 echo -n " -c")

    if [ "${config_files[*]}" == "-c" ]; then
        fail "No config files found. Please run '$0 configure'"
    fi
    if is_offline || "${no_pull}"; then
        warn "Pulling images is disabled by --no-pull or CBW_NO_PULL=true"
        extra_args=("--resolve-image" "never")
    else
        extra_args=("--with-registry-auth")
    fi

    CBW_ORCHESTRATOR_VERSION=$(docker version -f "swarm {{.Server}}" | tr '{' '(' | tr '}' ')')
    export CBW_ORCHESTRATOR_VERSION

    # Check that the docker version is compatible with the --detach option (26+). If not, don't use the option.
    docker_version=$(docker version -f "{{.Server.Version}}")
    detach_option=$({ printf "26.0.0\n%s\n" "$docker_version" | sort -C -V ; } && echo "--detach" || echo "" )

    # If present, remove the network for container-scanner. In some cases, this network is in an
    # error state and cannot be used by sidekiq_node. The container sidekiq_node is then in error
    # and cannot starts. The solution is to remove and create again the network.
    if docker network inspect "${network_container_scanner}" >/dev/null 2>&1; then
        echo "Network ${network_container_scanner} exists. Remove it."
        docker network rm ${network_container_scanner} || warn "Cannot remove network ${network_container_scanner}"
    fi

    # shellcheck disable=SC1090,SC2086
    (set -a && source "${containers_file}" && set +a &&
        docker stack deploy ${detach_option} "${config_files[@]}" "${extra_args[@]}" "$swarm_stack_name") ||
        fail "Cannot start docker swarm containers"

    read -ra docker_compose_config_files < <(find "${enabled_config_dir}/compose-service" -maxdepth 1 -name "*.yml" -print0 | sort -z | xargs -0 -n 1 echo -n " -f")
    # Start compose services after starting the swarm stack.
    # container-scanner needs the network created by the sidekiq_node compose file.
    # shellcheck disable=SC2015
    docker_compose_starts=$(docker compose "${docker_compose_config_files[@]}" up -d 2>&1) &&
        echo "Compose service container created" ||
        warn "Cannot starts one or multiple Compose service container(s). Reason:
'${docker_compose_starts}'"

    success "All Containers successfully deployed"
    prune_unused_images
}

cbw_stop() {

    # In all cases, try to stop docker compose. In some cases, the swarm stack can be down but not the docker compose containers.
    read -ra docker_compose_config_files < <(find "${enabled_config_dir}/compose-service" -maxdepth 1 -name "*.yml" -print0 | sort -z | xargs -0 -n 1 echo -n " -f")
    echo "Stopping Compose service containers ..."
    # shellcheck disable=SC2015
    docker_compose_stop=$(docker compose "${docker_compose_config_files[@]}" down --volumes 2>&1) &&
        echo "Compose service containers stopped" ||
            warn "Cannot stop Compose service containers. Reason:
                '${docker_compose_stop}'"

    is_running || {
        success "Cyberwatch On Premise is already stopped"
        return
    }

    stop_instance

    info "Cyberwatch On Premise is currently started, stopping"

    success "Cyberwatch On Premise successfully stopped"
}

stop_instance() {
    # Stop the swarm stack
    if ! docker stack rm "$swarm_stack_name"; then fail "Cannot stop docker swarm containers. Please reboot the virtual machine"; fi

    TIMEOUT=60  # 1m
    loop_start_time=$(date +%s)

    default_network_down=0
    container_scanner_network_down=0

    echo -n "Wait for docker."

    while [ $((default_network_down + container_scanner_network_down)) -ne 2 ]; do
        if docker events --since 10s --until 0s --filter 'type=network' --filter 'event=destroy' | grep "${swarm_stack_name}_default" >/dev/null; then
            default_network_down=1
        fi
        if docker events --since 10s --until 0s --filter 'type=network' --filter 'event=destroy' | grep "${network_container_scanner}" >/dev/null; then
            container_scanner_network_down=1
        fi
        echo -n "."
        sleep 1
        # Loop during TIMEOUT seconds before raising an error.
        [ $(($(date +%s) - loop_start_time)) -gt $TIMEOUT ] && echo "" && error "Timeout. Cannot handle the docker network destroy events for '${swarm_stack_name}_default' or '${network_container_scanner}'" && return
    done
    echo ""
}

cbw_restart() {
    LONG=skip-backup,no-pull,verify
    OPTS=$(getopt --longoptions $LONG -- "$0" "$@")

    eval set -- "$OPTS"

    while true; do
        case "$1" in
            "--skip-backup")
                export skip_backup=true
                shift
                ;;
            "--no-pull")
                no_pull=true
                shift
                ;;
            "--verify")
                verify=true
                shift
                ;;
            --)
                shift;
                break
                ;;
            *)
                echo "Unexpected option: $1"
            ;;
        esac
    done

    # Function used to migrate file for new version of cbw-on-premise.
    migrate_version

    [ -f "$containers_file" ] || fail "File $containers_file does not exist. Did you run '$0 configure' ?"

    is_satellite || backup_db_if_no_recent
    get_last_backup
    [ -d "${sshkey_dir}" ] || configure_ssh_key

    if is_offline || "${no_pull}"; then
        warn "Pulling images is disabled by --no-pull or CBW_NO_PULL=true"
    else
        synchronous_images_pull
    fi
    if ! is_offline && [ "$verify" = "true" ]; then
        cbw_verify-images-signature || fail "Failed to verify the signature of images"
        success "The signature of the images is valid."
    fi

    # The config may not be enabled if the instances was configured for the last time with version 5.8 of cbw-on-premise.
    # Let's ensure a restart of the instance enables it.
    true; enable_config 35-web-scanner.yml

    cbw_stop
    cbw_start
}

load_secrets() {
    # Delete the secrets before recreating them with updated values
    docker secret rm mysql-password mysql-root-password redis-password olympe >/dev/null 2>&1

    # The MariaDB container can read password from a file. To do that, each file
    # must contains only the value.
    use_external_db || extract_env_value MYSQL_ROOT_PASSWORD "$secret_file" | docker secret create mysql-root-password - >/dev/null
    extract_env_value MYSQL_PASSWORD      "$secret_file" | docker secret create mysql-password - >/dev/null

    # The redis does not support reading the password from file, we need to pass it as argument.
    # shellcheck disable=SC2046
    export $(grep REDIS_PASSWORD "$secret_file")

    # The MYSQL_ROOT_PASSWORD variable can be empty or unset. Docker does not
    # support empty secrets. The variable is transmitted to the sidekiq
    # container via environment variable.
    mysql_root_password_line=$(grep -vE ^# "$secret_file" | grep MYSQL_ROOT_PASSWORD)
    [ -n "$mysql_root_password_line" ] && export "${mysql_root_password_line?}"

    {
        grep -e REDIS_PASSWORD -e MYSQL_PASSWORD -e SECRET_KEY_BASE -e SECRET_KEY_CREDENTIAL "$secret_file";
        echo "LOCAL_API_TOKEN=$(cbw_get-local-api-token)";
        grep -e CBW_NODE_NAME "$containers_file"
     } | docker secret create olympe - >/dev/null
}

migrate_compose() {
    # Search for all config files containing the 'version' field.
    readarray -t config_files < <(grep -Rl -E "^version: .*" "${enabled_config_dir}" --include \*.yml)

    # For each config file with 'version' field.
    # Store a backup of the file 'in case of' and remove the field in place.
    for config_file in "${config_files[@]}"; do
        config_file_old="${config_file}.old"
        warn "Config file '$config_file' contains a compose 'version' field that is obsolete. The field will be removed!
A backup of the config file is available in '$config_file_old'"
        cp "$config_file" "$config_file_old"
        sed -i -E "s/^version: .*//" "$config_file"
    done
}

load_ssh_secret() {
    docker secret rm web-scanner-ssh-authorized-keys ssh-private-key >/dev/null 2>&1
    docker secret create ssh-private-key "${sshkey_dir}/id_ed25519" >/dev/null 2>&1
    docker secret create web-scanner-ssh-authorized-keys "${sshkey_dir}/id_ed25519.pub" >/dev/null 2>&1
}

extract_env_value() {
    variable=$1
    file=$2
    sed -n "s/^$variable=//p" "$file" | tail -n 1
}

cbw_status() {
    echo "SERVICES"
    docker stack services "$swarm_stack_name"
    echo
    echo "CONTAINERS"
    docker stack ps "$swarm_stack_name"
    echo
    echo "CONTAINERS COMPOSE"
    docker compose -f "${available_config_dir}/compose-service/container-scanner.yml" ps --format 'table {{.ID}}\t{{.Name}}\t{{.Image}}\t{{.Status}}\t{{.Ports}}'
}

synchronous_images_pull() {
    if docker compose version >/dev/null; then
        success "docker compose is installed, pulling containers..."

        read -ra config_files < <(find "${enabled_config_dir}" -name "*.yml" -print0 | sort -z | xargs -0 -n 1 echo -n " -f")

        if [ "${config_files[*]}" == "-f" ]; then
            fail "No config files found. Please run '$0 configure'"
        fi
        docker compose "${config_files[@]}" --env-file "${containers_file}" pull
        docker compose -f "${available_config_dir}/compose-service/container-scanner.yml" --env-file "${containers_file}" pull
    fi
}

## DEBUG

cbw_logs() {
    if [ -z "$1" ]; then
        error "Please specify a service name."
        cbw_help
        return 1
    fi
    CONTAINER="$1"
    if [[ "${CONTAINER}" == "container-scanner" ]]; then
       shift
       docker compose -f "${available_config_dir}/compose-service/container-scanner.yml" logs container-scanner -t "$@"
       exit 0
    fi
    shift
    docker service logs "${swarm_stack_name}_${CONTAINER}" --raw "$@"
}

cbw_exec() {
    if [ -z "$1" ]; then
        error "Please specify a service name"
        cbw_help
        return 1
    fi
    ARGS="${*:2}"
    if [[ "${1}" == "container-scanner" ]]; then
       shift
       docker compose -f "${available_config_dir}/compose-service/container-scanner.yml" exec container-scanner sh -c "${ARGS:-/bin/sh}"
       exit 0
    fi
    CONTAINER=$(docker ps -f name="${swarm_stack_name}_${1}\." -q | head -1)
    if [ -z "${CONTAINER}" ]; then
        warn "No such container: $1"
        exit 1
    fi
    shift
    if [ "$exec_interactive" == "false" ]; then
        docker exec -i "${CONTAINER}" sh -c "${ARGS:-/bin/sh}"
    else
        docker exec -it "${CONTAINER}" sh -c "${ARGS:-/bin/sh}"
    fi
}

cbw_show-config() {
    read -ra config_files < <(find "${enabled_config_dir}" -name "*.yml" -print0 | sort -z | xargs -0 -n 1 echo -n " -f")

    if [ "${config_files[*]}" == "-f" ]; then
        fail "No config files found. Please run '$0 configure'"
    fi

    set -a
    # shellcheck disable=SC1090
    source "${containers_file}"
    set +a
    docker compose "${config_files[@]}" config
}

check_orchestrator() {
    if command -v docker >/dev/null; then
        success "  > Docker version $(docker version --format '{{.Server.Version}}')"
        if docker compose version >/dev/null 2>&1; then
            success "    > $(docker compose version)"
        else
            warn "     > Docker plugin 'compose' is not installed"
        fi
    fi
}

check_container_status() {
    containers=$(
        docker ps \
        --format 'table {{ .Label "com.docker.swarm.service.name" }}\t{{ .Image }}\t{{ .State }}\t{{ .Status }}\t' \
        --filter "label=com.docker.stack.namespace=$swarm_stack_name" \
        --last 12 \
            | sed '1d'
                          )
    if echo "$containers" | grep -q 'running'; then
        if echo "$containers" | grep -q -v 'running'; then
            error "  > Found non running container(s):"
            echo -e "\n$containers\n" | grep -v 'running'
        else
            success "  > All container are running"
        fi
    else
        error "  > Cyberwatch is started but no running container found"
        echo
        docker stack ps cbwonpremise --format '{{ .Error }}' --no-trunc | grep -v '^$' | sort | uniq
        echo
    fi
}

report_orchestrator_config() {
    report_file_content "$config_dir/containers.env"
    for file in "$config_dir"/configs-enabled/*.yml; do
        report_file_content "$file"
    done
}

report_orchestrator_version() {
    report_command "docker version"
    report_command "docker compose version"
}

report_orchestrator_containers() {
    if is_running; then
        for service in $(docker stack services cbwonpremise -q --format '{{ .Name }}'); do
            report_command "docker inspect $service"
            report_command "docker service logs $service 2>&1"
        done
    fi
    report_command "docker images"
}

## MAINTENANCE

prune_unused_images() {
    [ -n "$CBW_DONT_PRUNE_UNUSED_IMAGES" ] && return
    if is_offline || "${no_pull}"; then
        if ! docker system prune -f 1>/dev/null; then fail "Cannot delete old files"; fi
    else
        if ! docker system prune -f -a 1>/dev/null; then fail "Cannot delete old files"; fi
    fi
    success "Successfully removed old docker files"
}

# Utils

is_satellite() { test -f "${enabled_config_dir}/15-satellite.yml"; }
use_external_db() { ! test -f "${enabled_config_dir}/40-db.yml"; }
is_offline() { [ "$is_offline" == "true" ] || [ -f "${enabled_config_dir}/70-offline.yml" ]; }
is_running() { docker stack ls | grep -q "$swarm_stack_name"; }
